@@ -1,10 +1,8 @@
 """EB  for the unit level model.
 
-This module implements the basic EB unit level model. The functionalities are organized in 
-classes. Each class has three main methods: *fit()*, *predict()* and *bootstrap_mse()*. Linear 
-Mixed Models (LMM) are the core underlying statistical framework used to model the hierarchical 
-nature of the small area estimation (SAE) techniques implemented in this module, 
-see McCulloch, C.E. and Searle, S.R. (2001) [#ms2001]_ for more details on LMM.
+This module implements the basic EB unit level model. The functionalities are organized in
+classes. Each class has three main methods: *fit()*, *predict()* and *bootstrap_mse()*. 
+Linear Mixed Models (LMM) are the core underlying statistical framework used to model the hierarchical nature of the small area estimation (SAE) techniques implemented in this module, see McCulloch, C.E. and Searle, S.R. (2001) [#ms2001]_ for more details on LMM.
 
 The *EbUnitModel* class implements the model developed by Molina, I. and Rao, J.N.K. (2010)
 [#mr2010]_. So far, only the basic approach requiring the normal distribution of the errors is 
@@ -24,146 +22,146 @@ see Rao, J.N.K. and Molina, I. (2015) [#rm2015]_.
    John Wiley & Sons, Hoboken, New Jersey.
 """
 
-from typing import Any, Dict, List, Optional, Tuple, Union, Callable
+from __future__ import annotations
+from typing import Any, Optional, Union, Callable
 
 import warnings
-import math
 
 import numpy as np
 import pandas as pd
-
 import statsmodels.api as sm
-
-from scipy.stats import norm as normal
-
-from samplics.utils import checks, formats, basic_functions
-from samplics.utils.types import Array, Number, StringNumber, DictStrNum
-
-from samplics.sae.sae_core_functions import area_stats
 from samplics.sae.eblup_unit_model import EblupUnitModel
+from samplics.utils import formats, basic_functions
+from samplics.utils.types import Array, DictStrNum, Number
 
 
 class EbUnitModel:
     """*EbUnitModel* implements the basic Unit level model for complex indicators.
 
-    *EbUnitModel* takes the sample data as input and fits the basic linear mixed model. 
-    The user can pick between restricted maximum likelihood (REML) or maximum likelihood (ML) 
-    to fit the model parameters. Also, EbUnitModel predicts the areas means and provides 
-    the point and mean squared error (MSE) estimates of the empirical Bayes linear 
+    *EbUnitModel* takes the sample data as input and fits the basic linear mixed model.
+    The user can pick between restricted maximum likelihood (REML) or maximum likelihood (ML)
+    to fit the model parameters. Also, EbUnitModel predicts the areas means and provides
+    the point and mean squared error (MSE) estimates of the empirical Bayes linear
     unbiased (EBLUP). User can also obtain the bootstrap mse estimates of the MSE.
 
-    *EbUnitModel* requires the user to provide the indicator function. The indicator function is 
-    expected to take the array of output sample observations as input and possibly some additional 
+    *EbUnitModel* requires the user to provide the indicator function. The indicator function is
+    expected to take the array of output sample observations as input and possibly some additional
     parameters needed to compute the indicator. The indicator function outputs an aggregated value.
-    For example, the poverty gap indicator can have the following signature 
-    pov_gap(y: array, pov_line: float) -> float. If the indicator function different outputs by 
+    For example, the poverty gap indicator can have the following signature
+    pov_gap(y: array, pov_line: float) -> float. If the indicator function different outputs by
     area then the self.area_list can be used to incorporate different logics across areas.
 
-    Also, *EbUnitModel* can use Boxcox to transform the output sample values in order to reduce 
-    the asymmetry in the datawhen fitting the linear mixed model. 
+    Also, *EbUnitModel* can use Boxcox to transform the output sample values in order to reduce
+    the asymmetry in the datawhen fitting the linear mixed model.
 
     Setting attributes
-        | method (str): the fitting method of the model parameters which can take the possible 
-        |   values restricted maximum likelihood (REML) or maximum likelihood (ML). 
-        |   If not specified, "REML" is used as default.  
-        | indicator (function): a user defined function to compute the indicator. 
-        | boxcox (dict): contains the *lambda* parameter of the Boxcox and a constant for the 
-        | log-transformation of the Boxcox.  
-    
+        | method (str): the fitting method of the model parameters which can take the possible
+        |   values restricted maximum likelihood (REML) or maximum likelihood (ML).
+        |   If not specified, "REML" is used as default.
+        | indicator (function): a user defined function to compute the indicator.
+        | boxcox (dict): contains the *lambda* parameter of the Boxcox and a constant for the
+        | log-transformation of the Boxcox.
+
     Sample related attributes
-        | ys (array): the output sample observations. 
-        | Xs (ndarray): the auxiliary information. 
+        | ys (array): the output sample observations.
+        | Xs (ndarray): the auxiliary information.
         | scales (array): an array of scaling parameters for the unit levels errors.
         | afactors (array): sum of the inverse squared of scale.
-        | areas (array): the full vector of small areas from the sampled observations.  
+        | areas (array): the full vector of small areas from the sampled observations.
         | areas_list (array): the list of small areas from the sample data.
-        | samp_size (dict): the sample size per small areas from the sample. 
-        | ys_mean (array): sample area means of the output variable. 
+        | samp_size (dict): the sample size per small areas from the sample.
+        | ys_mean (array): sample area means of the output variable.
         | Xs_mean (ndarray): sample area means of the auxiliary variables.
 
     Model fitting attributes
-        | fitted (boolean): indicates whether the model has been fitted or not. 
-        | fixed_effects (array): the estimated fixed effects of the regression model. 
-        | fe_std (array): the estimated standard errors of the fixed effects. 
+        | fitted (boolean): indicates whether the model has been fitted or not.
+        | fixed_effects (array): the estimated fixed effects of the regression model.
+        | fe_std (array): the estimated standard errors of the fixed effects.
         | random_effects (array): the estimated area level random effects.
-        |   associated with the small areas. 
-        | re_std (number): the estimated standard error of the random effects. 
-        | error_std (number): standard error of the unit level residuals. 
-        | convergence (dict): a dictionnary holding the convergence status and the number of 
-        |   iterations from the model fitting algorithm. 
+        |   associated with the small areas.
+        | re_std (number): the estimated standard error of the random effects.
+        | error_std (number): standard error of the unit level residuals.
+        | convergence (dict): a dictionnary holding the convergence status and the number of
+        |   iterations from the model fitting algorithm.
         | goodness (dict): a dictionary holding the log-likelihood, AIC, and BIC.
-        | gamma (dict): ratio of the between-area variability (re_std**2) to the total 
-        |   variability (re_std**2 + error_std**2 / a_factor). 
+        | gamma (dict): ratio of the between-area variability (re_std**2) to the total
+        |   variability (re_std**2 + error_std**2 / a_factor).
 
     Prediction related attributes
-        | areap (array): the list of areas for the prediction. 
-        | number_reps (int): number of replicates for the bootstrap MSE estimation. 
-        | area_est (array): area level EBLUP estimates. 
-        | area_mse (array): area level taylor estimation of the MSE. 
+        | areap (array): the list of areas for the prediction.
+        | number_reps (int): number of replicates for the bootstrap MSE estimation.
+        | area_est (array): area level EBLUP estimates.
+        | area_mse (array): area level taylor estimation of the MSE.
         | area_mse_boot (array): area level bootstrap estimation of the MSE.
 
     Main methods
         | fit(): fits the linear mixed model to estimate the model parameters using REMl or ML
-        |   methods. 
-        | predict(): predicts the area level indicator estimates which includes both the point 
-        |   estimates and the taylor MSE estimate. 
+        |   methods.
+        | predict(): predicts the area level indicator estimates which includes both the point
+        |   estimates and the taylor MSE estimate.
         | bootstrap_mse(): computes the area level bootstrap MSE estimates of the indicator.
     """
 
     def __init__(
-        self, method: str = "REML", boxcox: Optional[float] = None, constant: Number = 0,
+        self,
+        method: str = "REML",
+        boxcox: Optional[Number] = None,
+        constant: Optional[Number] = None,
     ):
 
         # Setting
         self.method: str = method.upper()
         if self.method not in ("REML", "ML"):
             raise AssertionError("Value provided for method is not valid!")
-        self.indicator: Callable[..., Any] = None
-        self.number_samples: Optional[int] = None
-        self.boxcox = {"lambda": boxcox, "constant": constant}
+        self.indicator: Callable[..., Any]
+        self.number_samples: int
+        self.boxcox: dict[str, Optional[Number]] = {"lambda": boxcox, "constant": constant}
 
         # Sample data
-        self.scales: np.ndarray = np.array([])
-        self.afactors: Dict[Any, float] = {}
-        self.ys: np.ndarray = np.array([])
-        self.Xs: np.ndarray = np.array([])
-        self.areas: np.ndarray = np.array([])
-        self.areas_list: np.ndarray = np.array([])
-        self.samp_size: Dict[Any, int] = {}
-        self.ys_mean: np.ndarray = np.array([])
-        self.Xs_mean: np.ndarray = np.array([])
+        self.scales: np.ndarray
+        self.afactors: DictStrNum
+        self.ys: np.ndarray
+        self.Xs: np.ndarray
+        self.areas: np.ndarray
+        self.areas_list: np.ndarray
+        self.samp_size: DictStrNum
+        self.ys_mean: np.ndarray
+        self.Xs_mean: np.ndarray
 
         # Fitted data
         self.fitted: bool = False
-        self.fixed_effects: np.ndarray = np.array([])
-        self.fe_std: np.ndarray = np.array([])
-        self.random_effects: np.ndarray = np.array([])
-        self.re_std: float = 0
-        self.error_std: float = 0
-        self.convergence: Dict[str, Union[float, int, bool]] = {}
-        self.goodness: Dict[str, float] = {}  # loglikehood, deviance, AIC, BIC
-        self.gamma: Dict[Any, float] = {}
+        self.fixed_effects: np.ndarray
+        self.fe_std: np.ndarray
+        self.random_effects: np.ndarray
+        self.re_std: float
+        self.error_std: float
+        self.convergence: dict[str, Union[float, int, bool]] = {}
+        self.goodness: dict[str, Number] = {}  # loglikehood, deviance, AIC, BIC
+        self.gamma: DictStrNum
 
         # Predict(ion/ed) data
-        self.number_reps: int = 0
-        self.area_est: Dict[Any, float] = {}
-        # self.area_mse: Dict[Any, float] = {}
-        self.area_mse_boot: Optional[Dict[Any, float]] = None
+        self.number_reps: int
+        self.area_est: DictStrNum
+        self.area_mse: DictStrNum
+        self.area_mse_boot: Optional[DictStrNum] = None
 
     def _transformation(self, y: np.ndarray, inverse: bool) -> np.ndarray:
         if self.boxcox["lambda"] is None:
-            z = y
-        elif self.boxcox["lambda"] == 0.0:
+            return y
+        elif self.boxcox["lambda"] == 0.0 and self.boxcox["constant"] is not None:
             if inverse:
-                z = np.exp(y) - self.boxcox["constant"]
+                return np.asarray(np.exp(y) - self.boxcox["constant"])
             else:
-                z = np.log(y + self.boxcox["constant"])
+                return np.asarray(np.log(y + self.boxcox["constant"]))
         elif self.boxcox["lambda"] != 0.0:
             if inverse:
-                z = np.exp(np.log(1 + y * self.boxcox["lambda"]) / self.boxcox["lambda"])
+                return np.asarray(
+                    np.exp(np.log(1 + y * self.boxcox["lambda"]) / self.boxcox["lambda"])
+                )
             else:
-                z = np.power(y, self.boxcox["lambda"]) / self.boxcox["lambda"]
-        return z
+                return np.asarray(np.power(y, self.boxcox["lambda"]) / self.boxcox["lambda"])
+        else:
+            raise AssertionError
 
     def fit(
         self,
@@ -177,30 +175,37 @@ class EbUnitModel:
         maxiter: int = 100,
     ) -> None:
         """Fits the linear mixed models to estimate the model parameters that is the fixed
-        effects, the random effects standard error and the unit level residuals' standard error. 
-        In addition, the method provides statistics related to the model fitting e.g. convergence 
+        effects, the random effects standard error and the unit level residuals' standard error.
+        In addition, the method provides statistics related to the model fitting e.g. convergence
         status, log-likelihood, AIC, BIC, and more.
 
         Args:
-            ys (Array): An array of the output sample observations. 
-            Xs (Array): An multi-dimensional array of the sample auxiliary information. 
-            areas (Array): provides the area of the sampled observations. 
-            samp_weight (Optional[Array], optional): An array of the sample weights. 
+            ys (Array): An array of the output sample observations.
+            Xs (Array): An multi-dimensional array of the sample auxiliary information.
+            areas (Array): provides the area of the sampled observations.
+            samp_weight (Optional[Array], optional): An array of the sample weights.
                 Defaults to None.
-            scales (Union[Array, Number], optional): the scale factor for the unit level errors. 
+            scales (Union[Array, Number], optional): the scale factor for the unit level errors.
                 If a single number of provided, the same number will be applied to all observations. Defaults to 1.
-            intercept (bool, optional): An boolean to indicate whether an intercept need to be 
+            intercept (bool, optional): An boolean to indicate whether an intercept need to be
                 added to Xs. Defaults to True
             tol (float, optional): tolerance used for convergence criteria. Defaults to 1.0e-4.
-            maxiter (int, optional): maximum number of iterations for the fitting algorithm. 
+            maxiter (int, optional): maximum number of iterations for the fitting algorithm.
             Defaults to 100.
         """
 
+        ys = formats.numpy_array(ys)
+
         ys_transformed = basic_functions.transform(
-            ys, llambda=self.boxcox["lambda"], constant=self.boxcox["constant"], inverse=False,
+            ys,
+            llambda=self.boxcox["lambda"],
+            constant=self.boxcox["constant"],
+            inverse=False,
         )
 
-        eblup_ul = EblupUnitModel(method=self.method,)
+        eblup_ul = EblupUnitModel(
+            method=self.method,
+        )
         eblup_ul.fit(
             ys_transformed, Xs, areas, samp_weight, scales, intercept, tol=tol, maxiter=maxiter
         )
@@ -250,16 +255,18 @@ class EbUnitModel:
                 Xs_mean = np.insert(self.Xs_mean.reshape(n, 1), 0, 1, axis=1)
             else:
                 Xs_mean = np.insert(self.Xs_mean, 0, 1, axis=1)
+        else:
+            Xs_mean = self.Xs_mean
 
         nb_arear = len(arear_list)
         mu_r = X_r @ fixed_effects
 
         if show_progress:
-            k = 0
             bar_length = min(50, nb_arear)
             steps = np.linspace(1, nb_arear - 1, bar_length).astype(int)
             print(f"Generating the {number_samples} replicates samples")
 
+        k = 0
         eta = np.zeros((number_samples, nb_arear)) * np.nan
         for i, d in enumerate(arear_list):
             # print(d)
@@ -275,11 +282,13 @@ class EbUnitModel:
             number_cycles = int(number_samples // cycle_size)
             last_cycle_size = number_samples % cycle_size
 
+            y_dr = None
             for j in range(number_cycles + 1):
                 if j == number_cycles:
                     cycle_size = last_cycle_size
                 re_effects = np.random.normal(
-                    scale=(sigma2u * (1 - self.gamma[d])) ** 0.5, size=cycle_size,
+                    scale=(sigma2u * (1 - self.gamma[d])) ** 0.5,
+                    size=cycle_size,
                 )
                 errors = np.random.normal(
                     scale=scale_dr * (sigma2e ** 0.5), size=(cycle_size, N_dr)
@@ -301,14 +310,17 @@ class EbUnitModel:
 
             y_d = np.append(y_dr, np.tile(y_s[area_s == d], [number_samples, 1]), axis=1)
             z_d = basic_functions.transform(
-                y_d, llambda=self.boxcox["lambda"], constant=self.boxcox["constant"], inverse=True,
+                y_d,
+                llambda=self.boxcox["lambda"],
+                constant=self.boxcox["constant"],
+                inverse=True,
             )
             eta[:, i] = np.apply_along_axis(indicator, axis=1, arr=z_d, **kwargs)  # *)
 
         if show_progress:
             print("\n")
 
-        return np.mean(eta, axis=0)
+        return np.asarray(np.mean(eta, axis=0))
 
     def predict(
         self,
@@ -323,24 +335,24 @@ class EbUnitModel:
         **kwargs: Any,
     ) -> None:
         """Predicts the area level means and provides the taylor MSE estimation of the estimated
-        area means. 
+        area means.
 
         Args:
             Xr (Array): an multi-dimensional array of the out of sample auxiliary variables.
             arear (Array): provides the area of the out of sample units.
             indicator (Callable[..., Array]): a user defined function which computes the area level
-                indicators. The function should take y (output variable) as the first parameters, 
-                additional parameters can be used. Use ***kwargs* to transfer the additional 
+                indicators. The function should take y (output variable) as the first parameters,
+                additional parameters can be used. Use ***kwargs* to transfer the additional
                 parameters.
-            number_samples (int): number of replicates for the Monte-Carlo (MC) algorithm.  
-            scaler (Union[Array, Number], optional): the scale factor for the unit level errors. 
+            number_samples (int): number of replicates for the Monte-Carlo (MC) algorithm.
+            scaler (Union[Array, Number], optional): the scale factor for the unit level errors.
                 If a single number of provided, the same number will be applied to all observations. Defaults to 1.
-            intercept (bool, optional): An boolean to indicate whether an intercept need to be 
+            intercept (bool, optional): An boolean to indicate whether an intercept need to be
                 added to Xr. Defaults to True.
-            max_array_length (int, optional): controls the number of replicates to generate at 
-                the same time. This parameter helps with performance. The number can be reduce or 
+            max_array_length (int, optional): controls the number of replicates to generate at
+                the same time. This parameter helps with performance. The number can be reduce or
                 increase based on the user's computer RAM capacity. Defaults to int(100e6).
-            show_progress (bool, optional): shows a bar progress of the MC replicates 
+            show_progress (bool, optional): shows a bar progress of the MC replicates
                 calculations. Defaults to True.
 
         Raises:
@@ -354,13 +366,13 @@ class EbUnitModel:
 
         self.number_samples = int(number_samples)
 
-        if isinstance(scaler, (float, int)):
-            scaler = np.ones(Xr.shape[0]) * scaler
-        else:
-            scale = formats.numpy_array(scaler)
-        area = formats.numpy_array(arear)
-        self.arear_list = np.unique(arear)
         Xr = formats.numpy_array(Xr)
+        arear = formats.numpy_array(arear)
+        self.arear_list = np.unique(arear)
+        if isinstance(scaler, (float, int)):
+            scaler = np.asarray(np.ones(Xr.shape[0]) * scaler)
+        else:
+            scaler = formats.numpy_array(scaler)
         if intercept:
             if Xr.ndim == 1:
                 n = Xr.shape[0]
@@ -369,6 +381,8 @@ class EbUnitModel:
             else:
                 Xr = np.insert(Xr, 0, 1, axis=1)
                 Xs = np.insert(self.Xs, 0, 1, axis=1)
+        else:
+            Xs = self.Xs
 
         area_est = self._predict_indicator(
             self.number_samples,
@@ -379,7 +393,7 @@ class EbUnitModel:
             arear,
             self.arear_list,
             self.fixed_effects,
-            self.gamma,
+            np.asarray(list(self.gamma.values())),
             self.error_std ** 2,
             self.re_std ** 2,
             scaler,
@@ -416,10 +430,10 @@ class EbUnitModel:
             scaler (Union[Array, Number], optional): [description]. Defaults to 1.
             intercept (bool, optional): [description]. Defaults to True.
             tol (float, optional): tolerance used for convergence criteria. Defaults to 1.0e-4.
-            maxiter (int, optional): maximum number of iterations for the fitting algorithm. 
+            maxiter (int, optional): maximum number of iterations for the fitting algorithm.
             Defaults to 100.
             max_array_length (int, optional): [description]. Defaults to int(100e6).
-            show_progress (bool, optional): shows a bar progress of the bootstrap replicates 
+            show_progress (bool, optional): shows a bar progress of the bootstrap replicates
                 calculations. Defaults to True.
 
         """
@@ -436,6 +450,8 @@ class EbUnitModel:
             else:
                 X_r = np.insert(X_r, 0, 1, axis=1)
                 Xs = np.insert(self.Xs, 0, 1, axis=1)
+        else:
+            Xs = self.Xs
 
         if isinstance(scaler, (float, int)):
             scale_r = np.ones(X_r.shape[0]) * scaler
@@ -470,9 +486,9 @@ class EbUnitModel:
             scale_dict[d] = scale[indice_dict[d]]
             a_factor_dict[d] = self.afactors[d] if d in self.areas_list else 0
             sample_size_dict[d] = self.samp_size[d] if d in self.areas_list else 0
-            scale_s_dict[d] = scale_s[area_s == d] if d in self.areas_list else 0
+            scale_s_dict[d] = scale_s[area_ds] if d in self.areas_list else 0
             X_dict[d] = X[indice_dict[d]]
-            X_s_dict[d] = X_s[area_s == d]
+            X_s_dict[d] = X_s[area_ds]
 
         cycle_size = max(int(max_array_length // sum(N_d)), 1)
         number_cycles = int(number_reps // cycle_size)
@@ -485,7 +501,7 @@ class EbUnitModel:
 
         eta_pop_boot = np.zeros((number_reps, nb_areas_ps))
         eta_samp_boot = np.zeros((number_reps, nb_areas_ps))
-        y_samp_boot = np.zeros((number_reps, np.sum(list(sample_size_dict.values()))))
+        y_samp_boot = np.zeros((number_reps, int(np.sum(list(sample_size_dict.values())))))
         print(f"Generating the {number_reps} bootstrap replicate populations")
         for b in range(number_cycles):
             start = b * cycle_size
@@ -494,11 +510,13 @@ class EbUnitModel:
                 end = number_reps
                 cycle_size = last_cycle_size
 
+            yboot_s = None
             for i, d in enumerate(areas_ps):
                 aboot_factor[i] = a_factor_dict[d]
 
                 re_d = np.random.normal(
-                    scale=self.re_std * (1 - self.gamma[d]) ** 0.5, size=cycle_size,
+                    scale=self.re_std * (1 - self.gamma[d]) ** 0.5,
+                    size=cycle_size,
                 )
                 err_d = np.random.normal(
                     scale=self.error_std * scale_dict[d],
@@ -514,9 +532,9 @@ class EbUnitModel:
                 eta_pop_boot[start:end, i] = indicator(zboot_d, **kwargs)
 
                 if i == 0:
-                    yboot_s = yboot_d[:, -sample_size_dict[d] :]
+                    yboot_s = yboot_d[:, -int(sample_size_dict[d]) :]
                 else:
-                    yboot_s = np.append(yboot_s, yboot_d[:, -sample_size_dict[d] :], axis=1)
+                    yboot_s = np.append(yboot_s, yboot_d[:, -int(sample_size_dict[d]) :], axis=1)
 
                 if show_progress:
                     run_id = b * nb_areas_ps + i
@@ -598,20 +616,21 @@ class EbUnitModel:
 
         print("\n")
 
-        mse_boot = np.mean(np.power(eta_samp_boot - eta_pop_boot, 2), axis=0)
+        mse_boot = np.asarray(np.mean(np.power(eta_samp_boot - eta_pop_boot, 2), axis=0))
         self.area_mse_boot = dict(zip(self.arear_list, mse_boot))
 
     def to_dataframe(
-        self, col_names: List[str] = ["_area", "_estimate", "_mse_boot"],
+        self,
+        col_names: list[str] = ["_area", "_estimate", "_mse_boot"],
     ) -> pd.DataFrame:
         """Returns a pandas dataframe from dictionaries with same keys and one value per key.
 
         Args:
-            col_names (list, optional): list of string to be used for the dataframe columns names. 
+            col_names (list, optional): list of string to be used for the dataframe columns names.
                 Defaults to ["_area", "_estimate", "_mse_boot"].
 
         Returns:
-            [pd.DataFrame]: a pandas dataframe 
+            [pd.DataFrame]: a pandas dataframe
         """
 
         ncols = len(col_names)
