@@ -14,66 +14,61 @@ level.
    Association*, **74**, 269-277.
 """
 
-from typing import Any, Dict, List, Optional, Tuple, Union
-
-import numpy as np
-import pandas as pd
+from __future__ import annotations
+from typing import Any, Union
 
 import math
-
-import statsmodels.api as sm
-
-from scipy.stats import norm as normal
-
-from samplics.utils import checks, formats
-from samplics.utils.types import Array, Number, StringNumber, DictStrNum
+import numpy as np
+import pandas as pd
+from samplics.utils import formats
+from samplics.utils.types import Array, DictStrNum, Number
 
 
 class EblupAreaModel:
-    """*EblupAreaModel* implements the basic area level model. 
+    """*EblupAreaModel* implements the basic area level model.
 
-    *EblupAreaModel* takes the unstable survey estimates as input and model them to produce the 
-    small area estimates. The standard error of the unit level residuals are assumed known and 
-    usually correspond to the survey estimates standard error. In some cases, the survey standard 
+    *EblupAreaModel* takes the unstable survey estimates as input and model them to produce the
+    small area estimates. The standard error of the unit level residuals are assumed known and
+    usually correspond to the survey estimates standard error. In some cases, the survey standard
     errors (or variances) are modelled to reduce their variability and the modelled standard errors
     are them used for the computes the small area estimates. The user can pick between REML, ML and
-    FH to estimate the model parameters. EblupAreaModel provides both the area level point estimates and the taylor approximation of the MSEs. 
+    FH to estimate the model parameters. EblupAreaModel provides both the area level point estimates and the taylor approximation of the MSEs.
 
     Setting attributes
-        | method (str): the fitting method of the model parameters which can take the possible 
-        |   values restricted maximum likelihood (REML), maximum likelihood (ML), and 
-        |   Fay-Herriot (FH). If not specified, "REML" is used as default.  
-    
+        | method (str): the fitting method of the model parameters which can take the possible
+        |   values restricted maximum likelihood (REML), maximum likelihood (ML), and
+        |   Fay-Herriot (FH). If not specified, "REML" is used as default.
+
     Sample related attributes
-        | yhat (array): the survey output area level estimates. This is also referred to as the 
-        | direct estimates. 
-        | error_std (number): the estimated standard error of the direct estimates. 
-        | X (ndarray): the auxiliary information. 
+        | yhat (array): the survey output area level estimates. This is also referred to as the
+        | direct estimates.
+        | error_std (number): the estimated standard error of the direct estimates.
+        | X (ndarray): the auxiliary information.
         | b_const (array): an array of scaling parameters for the area random effects.
-        | area (array): the area of the survey output estimates.  
-        | samp_size (dict): the sample size per small areas from the sample. 
-        | ys_mean (array): sample area means of the output variable. 
+        | area (array): the area of the survey output estimates.
+        | samp_size (dict): the sample size per small areas from the sample.
+        | ys_mean (array): sample area means of the output variable.
         | Xs_mean (ndarray): sample area means of the auxiliary variables.
 
     Model fitting attributes
-        | fitted (boolean): indicates whether the model has been fitted or not. 
-        | fixed_effects (array): the estimated fixed effects of the regression model. 
-        | fe_std (array): the estimated standard errors of the fixed effects. 
+        | fitted (boolean): indicates whether the model has been fitted or not.
+        | fixed_effects (array): the estimated fixed effects of the regression model.
+        | fe_std (array): the estimated standard errors of the fixed effects.
         | random_effects (array): the estimated standard errors of the fixed effects.
         | re_std (number): the estimated area level standard error of the random effects.
-        | convergence (dict): a dictionnary holding the convergence status and the number of 
-        |   iterations from the model fitting algorithm. 
+        | convergence (dict): a dictionnary holding the convergence status and the number of
+        |   iterations from the model fitting algorithm.
         | goodness (dict): a dictionary holding the log-likelihood, AIC, and BIC.
 
     Prediction related attributes
-        | area_est (array): area level modelled estimates. 
-        | area_mse (array): area level taylor estimation of the MSE. 
+        | area_est (array): area level modelled estimates.
+        | area_mse (array): area level taylor estimation of the MSE.
 
     Main methods
         | fit(): fits the linear mixed model to estimate the model parameters using REMl or ML
-        |   methods. 
-        | predict(): predicts the area level mean estimates which includes both the point   | | 
-        |   estimates and the taylor MSE estimate. 
+        |   methods.
+        | predict(): predicts the area level mean estimates which includes both the point   | |
+        |   estimates and the taylor MSE estimate.
     """
 
     def __init__(self, method: str = "REML") -> None:
@@ -84,25 +79,25 @@ class EblupAreaModel:
             self.method = method.upper()
 
         # Sample data
-        self.yhat: np.ndarray = np.array([])
-        self.error_std: np.ndarray = np.array([])
-        self.X: np.ndarray = np.array([])
-        self.area: np.ndarray = np.array([])
+        self.yhat: np.ndarray
+        self.error_std: np.ndarray
+        self.X: np.ndarray
+        self.area: np.ndarray
 
         # Fitting stats
         self.fitted: bool = False
-        self.fixed_effects: np.ndarray = np.array([])
-        self.fe_std: np.ndarray = np.array([])
-        self.re_std: np.ndarray = np.array([])
-        self.convergence: Dict[str, Union[float, int, bool]] = {}
-        self.goodness: Dict[str, float] = {}  # loglikehood, deviance, AIC, BIC
+        self.fixed_effects: np.ndarray
+        self.fe_std: np.ndarray
+        self.re_std: Number
+        self.convergence: dict[str, Union[float, int, bool]] = {}
+        self.goodness: dict[str, Number] = {}  # loglikehood, deviance, AIC, BIC
 
         # Predict(ino/ed) data
-        self.area_est: Dict[Any, float] = {}
-        self.area_mse: Dict[Any, float] = {}
-        self.area_mse_as1: Dict[Any, float] = {}
-        self.area_mse_as2: Dict[Any, float] = {}
-        self.area_mse_terms: Dict[str, Dict[Any, float]] = {}
+        self.area_est: DictStrNum
+        self.area_mse: DictStrNum
+        self.area_mse_as1: DictStrNum
+        self.area_mse_as2: DictStrNum
+        self.area_mse_terms: dict[str, DictStrNum]
 
     def __str__(self) -> str:
 
@@ -128,7 +123,7 @@ class EblupAreaModel:
         sigma2_e: np.ndarray,
         sigma2_v: float,
         b_const: np.ndarray,
-    ) -> Tuple[np.ndarray, np.ndarray]:
+    ) -> tuple[np.ndarray, np.ndarray]:
 
         V = np.diag(sigma2_v * (b_const ** 2) + sigma2_e)
         V_inv = np.linalg.inv(V)
@@ -141,7 +136,7 @@ class EblupAreaModel:
 
     def _log_likelihood(
         self, y: np.ndarray, X: np.ndarray, beta: np.ndarray, V: np.ndarray
-    ) -> float:
+    ) -> Number:
 
         m = y.size
         const = m * np.log(2 * np.pi)
@@ -168,16 +163,21 @@ class EblupAreaModel:
         yhat: np.ndarray,
         X: np.ndarray,
         sigma2_e: np.ndarray,
-        sigma2_v: np.ndarray,
+        sigma2_v: Number,
         b_const: np.ndarray,
-    ) -> Tuple[float, float]:
+    ) -> tuple[Number, Number]:
 
+        deriv_sigma = 0.0
+        info_sigma = 0.0
         if self.method == "ML":
             beta, beta_cov = self._fixed_coefficients(
-                area=area, yhat=yhat, X=X, sigma2_e=sigma2_e, sigma2_v=sigma2_v, b_const=b_const,
+                area=area,
+                yhat=yhat,
+                X=X,
+                sigma2_e=sigma2_e,
+                sigma2_v=sigma2_v,
+                b_const=b_const,
             )
-            deriv_sigma = 0.0
-            info_sigma = 0.0
             for d in area:
                 b_d = b_const[area == d]
                 phi_d = sigma2_e[area == d]
@@ -200,16 +200,19 @@ class EblupAreaModel:
             P = v_inv - np.matmul(np.matmul(v_inv, x_xvinvx_x), v_inv)
             P_B = np.matmul(P, B)
             P_B_P = np.matmul(P_B, P)
-            term1 = np.trace(P_B)
+            term1 = float(np.trace(P_B))
             term2 = np.matmul(np.matmul(np.transpose(yhat), P_B_P), yhat)
             deriv_sigma = -0.5 * (term1 - term2)
-            info_sigma = 0.5 * np.trace(np.matmul(P_B_P, B))
+            info_sigma = 0.5 * float(np.trace(np.matmul(P_B_P, B)))
         elif self.method == "FH":  # Fay-Herriot approximation
             beta, beta_cov = self._fixed_coefficients(
-                area=area, yhat=yhat, X=X, sigma2_e=sigma2_e, sigma2_v=sigma2_v, b_const=b_const,
+                area=area,
+                yhat=yhat,
+                X=X,
+                sigma2_e=sigma2_e,
+                sigma2_v=sigma2_v,
+                b_const=b_const,
             )
-            deriv_sigma = 0.0
-            info_sigma = 0.0
             for d in area:
                 b_d = b_const[area == d]
                 phi_d = sigma2_e[area == d]
@@ -236,27 +239,29 @@ class EblupAreaModel:
         sigma2_v_start: float,
         tol: float,
         maxiter: int,
-    ) -> Tuple[float, float, int, float, bool]:  # May not need variance
-        """ Fisher-scroring algorithm for estimation of variance component"""
+    ) -> tuple[float, float, int, float, bool]:  # May not need variance
+        """Fisher-scroring algorithm for estimation of variance component
+        return (sigma, covariance, number_iterations, tolerance, covergence status)"""
 
         iterations = 0
         tolerance = tol + 1.0
         sigma2_v = sigma2_v_start
+        info_sigma = 0.0
         while iterations < maxiter and tolerance > tol:
             sigma2_v_previous = sigma2_v
             deriv_sigma, info_sigma = self._partial_derivatives(
-                area=area, yhat=yhat, X=X, sigma2_e=sigma2_e, sigma2_v=sigma2_v, b_const=b_const,
+                area=area,
+                yhat=yhat,
+                X=X,
+                sigma2_e=sigma2_e,
+                sigma2_v=sigma2_v,
+                b_const=b_const,
             )
             sigma2_v += deriv_sigma / info_sigma
-            sigma2_v_cov = 1 / info_sigma
-
             tolerance = abs(sigma2_v - sigma2_v_previous)
-            convergence = tolerance <= tol
             iterations += 1
 
-        sigma2_v = float(max(sigma2_v, 0))
-
-        return sigma2_v, sigma2_v_cov, iterations, tolerance, convergence
+        return float(max(sigma2_v, 0)), 1 / info_sigma, iterations, tolerance, tolerance <= tol
 
     def _eb_estimates(
         self,
@@ -264,10 +269,11 @@ class EblupAreaModel:
         beta: np.ndarray,
         area: np.ndarray,
         sigma2_e: np.ndarray,
-        sigma2_v: float,
-        sigma2_v_cov: float,
+        sigma2_v: Number,
+        sigma2_v_cov: Number,
         b_const: np.ndarray,
-    ) -> Tuple[
+    ) -> tuple[
+        np.ndarray,
         np.ndarray,
         np.ndarray,
         np.ndarray,
@@ -275,15 +281,11 @@ class EblupAreaModel:
         np.ndarray,
         np.ndarray,
         np.ndarray,
-        np.dtype,
     ]:
 
         m = self.yhat.size
         v_i = sigma2_e + sigma2_v * (b_const ** 2)
-        V = np.diag(v_i)
         V_inv = np.diag(1 / v_i)
-        mu = np.dot(X, beta)
-        resid = self.yhat - mu
         G = np.diag(np.ones(m) * sigma2_v)
 
         Z = np.diag(b_const)
@@ -296,7 +298,7 @@ class EblupAreaModel:
         b_term_ml1 = np.linalg.inv(x_vinv_x)
         b_term_ml2_diag = (b_const ** 2) / (v_i ** 2)
         b_term_ml2 = np.matmul(np.matmul(np.transpose(X), np.diag(b_term_ml2_diag)), X)
-        b_term_ml = np.trace(np.matmul(b_term_ml1, b_term_ml2))
+        b_term_ml = float(np.trace(np.matmul(b_term_ml1, b_term_ml2)))
 
         estimates = np.array(self.yhat) * np.nan
         g1 = np.array(self.yhat) * np.nan
@@ -307,16 +309,18 @@ class EblupAreaModel:
         g1_partial = np.array(self.yhat) * np.nan
 
         sum_inv_vi2 = np.sum(1 / (v_i ** 2))
+        b_sigma2_v = 0.0
         if self.method == "REML":
-            b_sigma2_v = 0
-            g3_scale = 2 / sum_inv_vi2
+            g3_scale = 2.0 / sum_inv_vi2
         elif self.method == "ML":
-            b_sigma2_v = -(1 / 2 * sigma2_v_cov) * b_term_ml
-            g3_scale = 2 / sum_inv_vi2
+            b_sigma2_v = -(1.0 / 2.0 * sigma2_v_cov) * b_term_ml
+            g3_scale = 2.0 / sum_inv_vi2
         elif self.method == "FH":
             sum_vi = np.sum((1 / v_i))
-            b_sigma2_v = 2 * (m * sum_inv_vi2 - sum_vi ** 2) / (sum_vi ** 3)
-            g3_scale = 2 * m / sum_vi ** 2
+            b_sigma2_v = 2.0 * (m * sum_inv_vi2 - sum_vi ** 2) / (sum_vi ** 3)
+            g3_scale = 2.0 * m / sum_vi ** 2
+        else:
+            g3_scale = 0.0
 
         for d in area:
             b_d = b_const[area == d]
@@ -335,6 +339,9 @@ class EblupAreaModel:
             g3_star[area == d] = (g3[area == d] / variance_d) * (resid_d ** 2)
             g1_partial[area == d] = (b_d ** 2) * ((1 - gamma_d) ** 2) * b_sigma2_v
 
+        mse = 0
+        mse1_area_specific = 0
+        mse2_area_specific = 0
         if self.method == "REML":
             mse = g1 + g2 + 2 * g3
             mse1_area_specific = g1 + g2 + 2 * g3_star
@@ -345,14 +352,14 @@ class EblupAreaModel:
             mse2_area_specific = g1 - g1_partial + g2 + g3 + g3_star
 
         return (
-            estimates,
-            mse,
-            mse1_area_specific,
-            mse2_area_specific,
-            g1,
-            g2,
-            g3,
-            g3_star,
+            np.asarray(estimates),
+            np.asarray(mse),
+            np.asarray(mse1_area_specific),
+            np.asarray(mse2_area_specific),
+            np.asarray(g1),
+            np.asarray(g2),
+            np.asarray(g3),
+            np.asarray(g3_star),
         )
 
     def fit(
@@ -362,37 +369,40 @@ class EblupAreaModel:
         area: Array,
         error_std: Array,
         re_std_start: float = 0.001,
-        b_const: Union[np.array, Number] = 1.0,
+        b_const: Union[np.ndarray, Number] = 1.0,
         intercept: bool = True,
         tol: float = 1e-8,
         maxiter: int = 100,
     ) -> None:
-        """Fits the linear mixed models to estimate the fixed effects and the standard error of 
-        the random effects. In addition, the method provides statistics related to the model 
-        fitting e.g. convergence status, log-likelihood, and more.  
+        """Fits the linear mixed models to estimate the fixed effects and the standard error of
+        the random effects. In addition, the method provides statistics related to the model
+        fitting e.g. convergence status, log-likelihood, and more.
 
         Args:
-            yhat (Array): an array of the estimated area level survey estimates also called 
-            the direct estimates. 
-            X (Array): an multi-dimensional array of the auxiliary information associated to 
-            the sampled areas. 
-            area (Array): provides the areas associated to the direct estimates. 
+            yhat (Array): an array of the estimated area level survey estimates also called
+            the direct estimates.
+            X (Array): an multi-dimensional array of the auxiliary information associated to
+            the sampled areas.
+            area (Array): provides the areas associated to the direct estimates.
             error_std (Array): [description]
             re_std_start (float, optional): [description]. Defaults to 0.001.
             b_const (Union[np.array, Number], optional): [description]. Defaults to 1.0.
             tol (float, optional): tolerance used for convergence criteria. Defaults to 1.0e-4.
-            maxiter (int, optional): maximum number of iterations for the fitting algorithm. 
+            maxiter (int, optional): maximum number of iterations for the fitting algorithm.
             Defaults to 100.
         """
 
-        if isinstance(b_const, (int, float)):
-            b_const = np.ones(area.size) * b_const
-        else:
-            b_const = formats.numpy_array(b_const)
-
         area = formats.numpy_array(area)
         yhat = formats.numpy_array(yhat)
         X = formats.numpy_array(X)
+
+        error_std = formats.numpy_array(error_std)
+        error_std = formats.numpy_array(error_std)
+        if isinstance(b_const, (int, float)):
+            b_const = np.asarray(np.ones(area.size) * b_const)
+        else:
+            b_const = formats.numpy_array(b_const)
+
         if intercept and isinstance(X, np.ndarray):
             X = np.insert(X, 0, 1, axis=1)
 
@@ -414,7 +424,12 @@ class EblupAreaModel:
         )
 
         beta, beta_cov = self._fixed_coefficients(
-            area=area, yhat=yhat, X=X, sigma2_e=error_std ** 2, sigma2_v=sigma2_v, b_const=b_const,
+            area=area,
+            yhat=yhat,
+            X=X,
+            sigma2_e=error_std ** 2,
+            sigma2_v=sigma2_v,
+            b_const=b_const,
         )
 
         self.yhat = yhat
@@ -442,36 +457,40 @@ class EblupAreaModel:
         self.fitted = True
 
     def predict(
-        self, X: Array, area: Array, b_const: Union[np.array, Number] = 1.0, intercept: bool = True
+        self,
+        X: Array,
+        area: Array,
+        b_const: Union[np.ndarray, Number] = 1.0,
+        intercept: bool = True,
     ) -> None:
-        """Provides the modelled area levels estimates and their MSE estimates. 
+        """Provides the modelled area levels estimates and their MSE estimates.
 
         Args:
-            X (Array): an multi-dimensional array of the auxiliary variables associated to 
-            areas to predict. 
-            area (Array): provides the areas for the prediction. 
-            error_std (Array): 
-            b_const (Union[np.array, Number], optional): [description]. Defaults to 1.0.
+            X (Array): an multi-dimensional array of the auxiliary variables associated to
+            areas to predict.
+            area (Array): provides the areas for the prediction.
+            error_std (Array):
+            b_const (Union[np.ndarray, Number], optional): [description]. Defaults to 1.0.
 
         Raises:
             Exception: [description]
         """
 
+        area = formats.numpy_array(area)
         if not self.fitted:
             raise Exception(
                 "The model must be fitted first with .fit() before running the prediction."
             )
 
-        if isinstance(b_const, (int, float)):
-            b_const = np.ones(area.size) * b_const
-        else:
-            b_const = formats.numpy_array(b_const)
-
-        area = formats.numpy_array(area)
         X = formats.numpy_array(X)
         if intercept and isinstance(X, np.ndarray):
             X = np.insert(X, 0, 1, axis=1)
 
+        if isinstance(b_const, (int, float)):
+            b_const = np.asarray(np.ones(area.size) * b_const)
+        else:
+            b_const = formats.numpy_array(b_const)
+
         point_est, mse, mse1, mse2, g1, g2, g3, g3_star = self._eb_estimates(
             X=X,
             area=area,
@@ -485,15 +504,15 @@ class EblupAreaModel:
         self.area_est = dict(zip(area, point_est))
         self.area_mse = dict(zip(area, mse))
 
-    def to_dataframe(self, col_names: List[str] = ["_area", "_estimate", "_mse"]) -> pd.DataFrame:
+    def to_dataframe(self, col_names: list[str] = ["_area", "_estimate", "_mse"]) -> pd.DataFrame:
         """Returns a pandas dataframe from dictionaries with same keys and one value per key.
 
         Args:
-            col_names (list, optional): list of string to be used for the dataframe columns names. 
+            col_names (list, optional): list of string to be used for the dataframe columns names.
                 Defaults to ["_area", "_estimate", "_mse"].
 
         Returns:
-            [type]: a pandas dataframe 
+            [type]: a pandas dataframe
         """
 
         return formats.dict_to_dataframe(col_names, self.area_est, self.area_mse)
