@@ -25,81 +25,77 @@ see Rao, J.N.K. and Molina, I. (2015) [#rm2015]_.
    John Wiley & Sons, Hoboken, New Jersey.
 """
 
-from typing import Any, Dict, List, Optional, Tuple, Union, Callable
+from __future__ import annotations
+from typing import Any, Optional, Union
 
 import warnings
-import math
 
 import numpy as np
 import pandas as pd
-
 import statsmodels.api as sm
-
-from scipy.stats import norm as normal
-
-from samplics.utils import checks, formats, basic_functions
-from samplics.utils.types import Array, Number, StringNumber, DictStrNum
-
 from samplics.sae.sae_core_functions import area_stats
+from samplics.utils import formats, basic_functions
+from samplics.utils.types import Array, DictStrNum, Number, StringNumber
 
 
 class EblupUnitModel:
     """*EblupUnitModel* implements the basic unit level model for means (a linear indicator).
 
-    *EblupUnitModel* takes the sample data as input and fits the basic linear mixed model. 
-    The user can pick between restricted maximum likelihood (REML) or maximum likelihood (ML) 
-    to fit the model parameters. Also, EblupUnitModel predicts the areas means and provides 
-    the point and mean squared error (MSE) estimates of the empirical Bayes linear 
+    *EblupUnitModel* takes the sample data as input and fits the basic linear mixed model.
+    The user can pick between restricted maximum likelihood (REML) or maximum likelihood (ML)
+    to fit the model parameters. Also, EblupUnitModel predicts the areas means and provides
+    the point and mean squared error (MSE) estimates of the empirical Bayes linear
     unbiased (EBLUP). User can also obtain the bootstrap mse estimates of the MSE.
 
     Setting attributes
-        | method (str): the fitting method of the model parameters which can take the possible 
-        |   values restricted maximum likelihood (REML) or maximum likelihood (ML). 
-        |   If not specified, "REML" is used as default.  
-    
+        | method (str): the fitting method of the model parameters which can take the possible
+        |   values restricted maximum likelihood (REML) or maximum likelihood (ML).
+        |   If not specified, "REML" is used as default.
+
     Sample related attributes
-        | ys (array): the output sample observations. 
-        | Xs (ndarray): the auxiliary information. 
+        | ys (array): the output sample observations.
+        | Xs (ndarray): the auxiliary information.
         | scales (array): an array of scaling parameters for the unit levels errors.
         | afactors (array): sum of the inverse squared of scale.
-        | areas (array): the full vector of small areas from the sampled observations.  
+        | areas (array): the full vector of small areas from the sampled observations.
         | areas_list (array): the list of small areas from the sample data.
-        | samp_size (dict): the sample size per small areas from the sample. 
-        | ys_mean (array): sample area means of the output variable. 
+        | samp_size (dict): the sample size per small areas from the sample.
+        | ys_mean (array): sample area means of the output variable.
         | Xs_mean (ndarray): sample area means of the auxiliary variables.}}}}}
 
     Model fitting attributes
-        | fitted (boolean): indicates whether the model has been fitted or not. 
-        | fixed_effects (array): the estimated fixed effects of the regression model. 
-        | fe_std (array): the estimated standard errors of the fixed effects. 
+        | fitted (boolean): indicates whether the model has been fitted or not.
+        | fixed_effects (array): the estimated fixed effects of the regression model.
+        | fe_std (array): the estimated standard errors of the fixed effects.
         | random_effects (array): the estimated area level random effects.
-        | re_std (number): the estimated standard error of the random effects. 
-        | error_std (number): the estimated standard error of the unit level residuals. 
-        | convergence (dict): a dictionnary holding the convergence status and the number of 
-        |   iterations from the model fitting algorithm. 
+        | re_std (number): the estimated standard error of the random effects.
+        | error_std (number): the estimated standard error of the unit level residuals.
+        | convergence (dict): a dictionnary holding the convergence status and the number of
+        |   iterations from the model fitting algorithm.
         | goodness (dict): a dictionary holding the log-likelihood, AIC, and BIC.
-        | gamma (dict): ratio of the between-area variability (re_std**2) to the total 
-        |   variability (re_std**2 + error_std**2 / a_factor). 
+        | gamma (dict): ratio of the between-area variability (re_std**2) to the total
+        |   variability (re_std**2 + error_std**2 / a_factor).
 
     Prediction related attributes
-        | areap (array): the list of areas for the prediction. 
-        | Xmean (array): population means of the auxiliary variables. 
-        | number_reps (int): number of replicates for the bootstrap MSE estimation. 
+        | areap (array): the list of areas for the prediction.
+        | Xmean (array): population means of the auxiliary variables.
+        | number_reps (int): number of replicates for the bootstrap MSE estimation.
         | samp_rate (dict): sampling rates at the area level.
-        | area_est (array): area level EBLUP estimates. 
-        | area_mse (array): area level taylor estimation of the MSE. 
+        | area_est (array): area level EBLUP estimates.
+        | area_mse (array): area level taylor estimation of the MSE.
         | area_mse_boot (array): area level bootstrap estimation of the MSE.
 
     Main methods
         | fit(): fits the linear mixed model to estimate the model parameters using REMl or ML
-        |   methods. 
-        | predict(): predicts the area level mean estimates which includes both the point  
-        |   estimates and the taylor MSE estimate. 
+        |   methods.
+        | predict(): predicts the area level mean estimates which includes both the point
+        |   estimates and the taylor MSE estimate.
         | bootstrap_mse(): computes the area level bootstrap MSE estimates of the mean.
     """
 
     def __init__(
-        self, method: str = "REML",
+        self,
+        method: str = "REML",
     ):
         # Setting
         self.method: str = method.upper()
@@ -107,38 +103,42 @@ class EblupUnitModel:
             raise AssertionError("Value provided for method is not valid!")
 
         # Sample data
-        self.scales: np.ndarray = np.array([])
-        self.afactors: Dict[Any, float] = {}
-        self.ys: np.ndarray = np.array([])
-        self.Xs: np.ndarray = np.array([])
-        self.areas: np.ndarray = np.array([])
-        self.areas_list: np.ndarray = np.array([])
-        self.samp_size: Dict[Any, int] = {}
-        self.ys_mean: np.ndarray = np.array([])
-        self.Xs_mean: np.ndarray = np.array([])
+        self.scales: np.ndarray 
+        self.afactors: DictStrNum
+        self.ys: np.ndarray 
+        self.Xs: np.ndarray
+        self.areas: np.ndarray 
+        self.areas_list: np.ndarray 
+        self.samp_size: DictStrNum
+        self.ys_mean: np.ndarray
+        self.Xs_mean: np.ndarray
 
         # Fitting stats
         self.fitted: bool = False
-        self.fixed_effects: np.ndarray = np.array([])
-        self.fe_std: np.ndarray = np.array([])
-        self.random_effects: np.ndarray = np.array([])
-        self.re_std: float = 0
-        self.error_std: float = 0
-        self.convergence: Dict[str, Union[float, int, bool]] = {}
-        self.goodness: Dict[str, float] = {}  # loglikehood, deviance, AIC, BIC
-        self.gamma: Dict[Any, float] = {}
+        self.fixed_effects: np.ndarray 
+        self.fe_std: np.ndarray 
+        self.random_effects: np.ndarray 
+        self.re_std: Number = 0
+        self.error_std: Number = 0
+        self.convergence: dict[str, Union[float, int, bool]] = {}
+        self.goodness: dict[str, Number] = {}  # loglikehood, deviance, AIC, BIC
+        self.gamma: DictStrNum
 
         # Predict(ion/ed) data
-        self.areap: np.ndarray = np.array([])
-        self.Xp_mean: np.ndarray = np.array([])
+        self.areap: np.ndarray
+        self.Xp_mean: np.ndarray
         self.number_reps: int = 0
-        self.samp_rate: Dict[Any, float] = None
-        self.area_est: Dict[Any, float] = None
-        self.area_mse: Dict[Any, float] = None
-        self.area_mse_boot: Optional[Dict[Any, float]] = None
+        self.samp_rate: DictStrNum
+        self.area_est: DictStrNum
+        self.area_mse: DictStrNum
+        self.area_mse_boot: Optional[DictStrNum] = None
 
     def _beta(
-        self, y: np.ndarray, X: np.ndarray, area: np.ndarray, weight: np.ndarray,
+        self,
+        y: np.ndarray,
+        X: np.ndarray,
+        area: np.ndarray,
+        weight: np.ndarray,
     ) -> np.ndarray:
 
         Xw = X * weight[:, None]
@@ -156,9 +156,8 @@ class EblupUnitModel:
             beta1 = beta1 + np.matmul(np.transpose(Xw_d), resid_d_w)
             beta2 = beta2 + np.sum(resid_d_w * y_d[:, None] * w_d[:, None], axis=0)
 
-        beta = np.matmul(np.linalg.inv(beta1), beta2)
-
-        return beta
+        return np.asarray(np.matmul(np.linalg.inv(beta1), beta2)
+)
 
     def _mse(
         self,
@@ -193,7 +192,7 @@ class EblupUnitModel:
             / i_determinant
         )
 
-        return g1 + g2 + 2 * g3
+        return np.asarray(g1 + g2 + 2 * g3)
 
     def fit(
         self,
@@ -207,22 +206,22 @@ class EblupUnitModel:
         maxiter: int = 100,
     ) -> None:
         """Fits the linear mixed models to estimate the model parameters that is the fixed
-        effects, the random effects standard error and the unit level residuals' standard error. 
-        In addition, the method provides statistics related to the model fitting e.g. convergence 
-        status, log-likelihood, AIC, BIC, and more.  
+        effects, the random effects standard error and the unit level residuals' standard error.
+        In addition, the method provides statistics related to the model fitting e.g. convergence
+        status, log-likelihood, AIC, BIC, and more.
 
         Args:
-            ys (Array): An array of the output sample observations. 
-            Xs (Array): An multi-dimensional array of the auxiliary information. 
-            areas (Array): An array of the sampled area provided at the unit level. 
-            samp_weight (Optional[Array], optional): An array of the sample weights. 
+            ys (Array): An array of the output sample observations.
+            Xs (Array): An multi-dimensional array of the auxiliary information.
+            areas (Array): An array of the sampled area provided at the unit level.
+            samp_weight (Optional[Array], optional): An array of the sample weights.
                 Defaults to None.
-            scales (Union[Array, Number], optional): The scale factor for the unit level errors. 
+            scales (Union[Array, Number], optional): The scale factor for the unit level errors.
                 If a single number of provided, the same number will be applied to all observations. Defaults to 1.
-            intercept (bool, optional): An boolean to indicate whether an intercept need to be 
+            intercept (bool, optional): An boolean to indicate whether an intercept need to be
                 added to X. Defaults to True.
             tol (float, optional): tolerance used for convergence criteria. Defaults to 1.0e-4.
-            maxiter (int, optional): maximum number of iterations for the fitting algorithm. 
+            maxiter (int, optional): maximum number of iterations for the fitting algorithm.
             Defaults to 100.
         """
 
@@ -241,13 +240,13 @@ class EblupUnitModel:
                 Xs = np.insert(Xs.reshape(n, 1), 0, 1, axis=1)
             else:
                 Xs = np.insert(Xs, 0, 1, axis=1)
-        if samp_weight is not None and isinstance(samp_weight, pd.DataFrame):
+        if samp_weight is not None:
             samp_weight = formats.numpy_array(samp_weight)
+
         if isinstance(scales, (float, int)):
-            scales = np.ones(ys.shape[0]) * scales
+            scales = np.asarray(np.ones(ys.shape[0]) * scales)
         else:
             scales = formats.numpy_array(scales)
-
         self.scales = scales
 
         self.afactors = dict(zip(self.areas_list, basic_functions.sumby(areas, scales)))
@@ -261,11 +260,9 @@ class EblupUnitModel:
             "maxiter": maxiter,
         }  # TODO: to improve in the future. Check: statsmodels.LikelihoodModel.fit()
         basic_fit = basic_model.fit(reml=reml, full_output=True, **fit_kwargs)
-        # print(basic_fit.summary())
 
         self.error_std = basic_fit.scale ** 0.5
         self.fixed_effects = basic_fit.fe_params
-        # self.random_effects = np.array(list(basic_fit.random_effects.values()))
 
         self.fe_std = basic_fit.bse_fe
         self.re_std = float(basic_fit.cov_re) ** 0.5
@@ -293,35 +290,41 @@ class EblupUnitModel:
         self.goodness["BIC"] = bic
 
         self.ys_mean, Xs_mean, gamma, samp_size = area_stats(
-            ys, Xs, areas, self.error_std, self.re_std, self.afactors, samp_weight,
+            ys,
+            Xs,
+            areas,
+            self.error_std,
+            self.re_std,
+            self.afactors,
+            samp_weight,
         )
         self.random_effects = gamma * (self.ys_mean - Xs_mean @ self.fixed_effects)
         self.gamma = dict(zip(self.areas_list, gamma))
         self.samp_size = dict(zip(self.areas_list, samp_size))
         self.Xs_mean = Xs_mean[:, 1:]
 
-        # samp_weight = np.ones(y.size)
-        if samp_weight is not None:
-            beta_w = self._beta(ys, Xs, areas, samp_weight)
-
         self.fitted = True
 
     def predict(
-        self, Xmean: Array, area: Array, pop_size: Optional[Array] = None, intercept: bool = True,
+        self,
+        Xmean: Array,
+        area: Array,
+        pop_size: Optional[Array] = None,
+        intercept: bool = True,
     ) -> None:
         """Predicts the area level means and provides the taylor MSE estimation of the estimated
-        area means. 
+        area means.
 
         Args:
-            Xmean (Array): a multi-dimensional array of the population means of the auxiliary   
-                variables. 
-            area (Array): An array of the areas in the same order as Xmean and Popsize. 
-            pop_size (Optional[Array], optional): An array of the population size for the same 
+            Xmean (Array): a multi-dimensional array of the population means of the auxiliary
+                variables.
+            area (Array): An array of the areas in the same order as Xmean and Popsize.
+            pop_size (Optional[Array], optional): An array of the population size for the same
                 areas as in Xmean and area and in the same order. Defaults to None.
             intercept (bool, optional): [description]. Defaults to True.
 
         Raises:
-            Exception: when predict() is called before fitting the model. 
+            Exception: when predict() is called before fitting the model.
         """
 
         if not self.fitted:
@@ -341,8 +344,13 @@ class EblupUnitModel:
                 Xp_mean = np.insert(Xmean, 0, 1, axis=1)
                 Xs_mean = np.insert(self.Xs_mean, 0, 1, axis=1)
                 Xs = np.insert(self.Xs, 0, 1, axis=1)
+        else:
+            Xp_mean = self.Xp_mean
+            Xs_mean = self.Xs_mean
+            Xs = self.Xs
 
         area = formats.numpy_array(area)
+        pop_size_dict = {}
         if pop_size is not None:
             pop_size = formats.numpy_array(pop_size)
             pop_size_dict = dict(zip(area, pop_size))
@@ -399,8 +407,8 @@ class EblupUnitModel:
         )
         self.area_mse = dict(zip(ps_area_list, mse_ps))
 
-        pr_area_list = self.areap[~ps]
         # TODO: add non-sampled areas prediction (section 7.2.2, Rao and molina (2015))
+        # pr_area_list = self.areap[~ps]
 
     def bootstrap_mse(
         self,
@@ -410,20 +418,20 @@ class EblupUnitModel:
         maxiter: int = 100,
         show_progress: bool = True,
     ) -> None:
-        """Computes the MSE bootstrap estimates of the area level mean estimates. 
+        """Computes the MSE bootstrap estimates of the area level mean estimates.
 
-            Args:
-                number_reps (int): Number of replicates for the bootstrap method.. Defaults to 500.
-                scale (Union[Array, Number], optional): [description]. Defaults to 1.
-                intercept (bool, optional): [description]. Defaults to True.
-                tol: float = 1.0e-4,
-                maxiter: int = 100,
-                show_progress (bool, optional): shows a bar progress of the bootstrap replicates 
-                    calculations. Defaults to True.
+        Args:
+            number_reps (int): Number of replicates for the bootstrap method.. Defaults to 500.
+            scale (Union[Array, Number], optional): [description]. Defaults to 1.
+            intercept (bool, optional): [description]. Defaults to True.
+            tol: float = 1.0e-4,
+            maxiter: int = 100,
+            show_progress (bool, optional): shows a bar progress of the bootstrap replicates
+                calculations. Defaults to True.
 
-            Raises:
-                Exception: when bootstrap_mse() is called before fitting the model. 
-            """
+        Raises:
+            Exception: when bootstrap_mse() is called before fitting the model.
+        """
 
         if not self.fitted:
             raise Exception(
@@ -434,17 +442,17 @@ class EblupUnitModel:
             if self.Xs.ndim == 1:
                 n = self.Xs.shape[0]
                 Xp_mean = np.insert(self.Xp_mean.reshape(n, 1), 0, 1, axis=1)
-                Xs_mean = np.insert(self.Xs_mean.reshape(n, 1), 0, 1, axis=1)
                 Xs = np.insert(self.Xs.reshape(n, 1), 0, 1, axis=1)
             else:
                 Xp_mean = np.insert(self.Xp_mean, 0, 1, axis=1)
-                Xs_mean = np.insert(self.Xs_mean, 0, 1, axis=1)
                 Xs = np.insert(self.Xs, 0, 1, axis=1)
+        else:
+            Xp_mean = self.Xp_mean
+            Xs = self.Xs
 
         nb_areas = self.areap.size
         ps = np.isin(self.areap, self.areas_list)
         ps_area_list = self.areap[ps]
-        ps_area_indices = np.isin(self.areas_list, ps_area_list)
 
         for i, d in enumerate(ps_area_list):
             aread = self.areas == d
@@ -453,7 +461,7 @@ class EblupUnitModel:
                 scale=self.error_std, size=(number_reps, self.samp_size[d])
             )
             re_boot_d = np.random.normal(scale=self.re_std, size=(number_reps, 1))
-            re_d = np.repeat(re_boot_d, self.samp_size[d], axis=1)
+            re_d = np.repeat(re_boot_d, int(self.samp_size[d]), axis=1)
 
             if i == 0:
                 error = error_d
@@ -485,14 +493,20 @@ class EblupUnitModel:
         boot_mse = np.zeros((number_reps, nb_areas))
         print(f"Running the {number_reps} bootstrap iterations")
         for b in range(number_reps):
-            with warnings.catch_warnings(record=True) as w:
+            with warnings.catch_warnings(record=True):
                 boot_model = sm.MixedLM(y_ps_boot[b, :], X_ps, area_ps)
                 boot_fit = boot_model.fit(reml=reml, **fit_kwargs)
             boot_fe = boot_fit.fe_params
             boot_error_std = boot_fit.scale ** 0.5
             boot_re_std = float(boot_fit.cov_re) ** 0.5
             boot_ys_mean, boot_Xs_mean, boot_gamma, _ = area_stats(
-                y_ps_boot[b, :], X_ps, area_ps, boot_error_std, boot_re_std, self.afactors, None,
+                y_ps_boot[b, :],
+                X_ps,
+                area_ps,
+                boot_error_std,
+                boot_re_std,
+                self.afactors,
+                None,
             )
             boot_re = boot_gamma * (boot_ys_mean - boot_Xs_mean @ boot_fe)
             boot_mu = Xp_mean @ self.fixed_effects + re_boot[b, :]
@@ -509,21 +523,22 @@ class EblupUnitModel:
         if show_progress:
             print("\n")
 
-        self.area_mse_boot = dict(zip(ps_area_list, np.mean(boot_mse, axis=0)))
+        self.area_mse_boot = dict(zip(ps_area_list, np.asarray(np.mean(boot_mse, axis=0))))
 
         # TODO: nonnegligeable sampling fractions, section 7.2.4, Rao and Molina (2015)
 
     def to_dataframe(
-        self, col_names: List[str] = ["_area", "_estimate", "_mse", "_mse_boot"],
+        self,
+        col_names: list[str] = ["_area", "_estimate", "_mse", "_mse_boot"],
     ) -> pd.DataFrame:
         """Returns a pandas dataframe from dictionaries with same keys and one value per key.
 
         Args:
-            col_names (list, optional): list of string to be used for the dataframe columns names. 
+            col_names (list, optional): list of string to be used for the dataframe columns names.
                 Defaults to ["_area", "_estimate", "_mse", "_mse_boot"].
 
         Returns:
-            [pd.DataFrame]: a pandas dataframe 
+            [pd.DataFrame]: a pandas dataframe
         """
 
         ncols = len(col_names)
