@@ -23,20 +23,16 @@ see Rao, J.N.K. and Molina, I. (2015) [#rm2015]_.
    John Wiley & Sons, Hoboken, New Jersey.
 """
 
-from typing import Any, Dict, List, Optional, Tuple, Union, Callable
-
-import warnings
-import math
+from __future__ import annotations
+from typing import Any, Optional, Union, Callable
 
 import numpy as np
 import pandas as pd
 
 import statsmodels.api as sm
 
-from scipy.stats import norm as normal
-
-from samplics.utils import checks, formats, basic_functions
-from samplics.utils.types import Array, Number, StringNumber, DictStrNum
+from samplics.utils import formats, basic_functions
+from samplics.utils.types import Array, DictStrNum, Number
 
 from samplics.sae.sae_core_functions import area_stats
 from samplics.sae.eb_unit_model import EbUnitModel
@@ -45,72 +41,72 @@ from samplics.sae.eb_unit_model import EbUnitModel
 class EllUnitModel:
     """*EllUnitModel* implements the basic Unit level model for complex indicators.
 
-    *EllUnitModel* takes the sample data as input and fits the basic linear mixed model. 
-    The user can pick between restricted maximum likelihood (REML), maximum likelihood (ML), 
-    and method of moments (MOM) to fit the model parameters. Also, EllUnitModel predicts the 
-    areas means and provides the point and mean squared error (MSE) estimates of the empirical 
-    Bayes linear unbiased (EBLUP). 
+    *EllUnitModel* takes the sample data as input and fits the basic linear mixed model.
+    The user can pick between restricted maximum likelihood (REML), maximum likelihood (ML),
+    and method of moments (MOM) to fit the model parameters. Also, EllUnitModel predicts the
+    areas means and provides the point and mean squared error (MSE) estimates of the empirical
+    Bayes linear unbiased (EBLUP).
 
-    *EllUnitModel* requires the user to provide the indicator function. The indicator function is 
-    expected to take the array of output sample observations as input and possibly some additional 
+    *EllUnitModel* requires the user to provide the indicator function. The indicator function is
+    expected to take the array of output sample observations as input and possibly some additional
     parameters needed to compute the indicator. The indicator function outputs an aggregated value.
-    For example, the poverty gap indicator can have the following signature 
-    pov_gap(y: array, pov_line: float) -> float. If the indicator function different outputs by 
+    For example, the poverty gap indicator can have the following signature
+    pov_gap(y: array, pov_line: float) -> float. If the indicator function different outputs by
     area then the self.area_list can be used to incorporate different logics across areas.
 
-    Also, EllUnitModel can use Boxcox to transform the output sample values in order to reduce the 
-    asymmetry in the datawhen fitting the linear mixed model. 
+    Also, EllUnitModel can use Boxcox to transform the output sample values in order to reduce the
+    asymmetry in the datawhen fitting the linear mixed model.
 
     Setting attributes
-        | method (str): the fitting method of the model parameters which can take the possible 
-        |   values restricted maximum likelihood (REML),  maximum likelihood (ML), and method of 
-        |   moments (MOM). If not specified, "MOM" is used as default.  
-        | indicator (function): a user defined function to compute the indicator. 
-        | boxcox (dict): contains the *lambda* parameter of the Boxcox and a constant for the 
-        |   log-transformation of the Boxcox.  
-    
+        | method (str): the fitting method of the model parameters which can take the possible
+        |   values restricted maximum likelihood (REML),  maximum likelihood (ML), and method of
+        |   moments (MOM). If not specified, "MOM" is used as default.
+        | indicator (function): a user defined function to compute the indicator.
+        | boxcox (dict): contains the *lambda* parameter of the Boxcox and a constant for the
+        |   log-transformation of the Boxcox.
+
     Sample related attributes
-        | ys (array): the output sample observations. 
-        | Xs (ndarray): the auxiliary information. 
+        | ys (array): the output sample observations.
+        | Xs (ndarray): the auxiliary information.
         | scales (array): an array of scaling parameters for the unit levels errors.
         | afactors (array): sum of the inverse squared of scale.
-        | areas (array): the full vector of small areas from the sampled observations.  
+        | areas (array): the full vector of small areas from the sampled observations.
         | areas_list (array): the list of small areas from the sample data.
-        | samp_size (dict): the sample size per small areas from the sample. 
-        | ys_mean (array): sample area means of the output variable. 
+        | samp_size (dict): the sample size per small areas from the sample.
+        | ys_mean (array): sample area means of the output variable.
         | Xs_mean (ndarray): sample area means of the auxiliary variables.
 
     Model fitting attributes
-        | fitted (boolean): indicates whether the model has been fitted or not. 
-        | fixed_effects (array): the estimated fixed effects of the regression model. 
-        | fe_std (array): the estimated standard errors of the fixed effects. 
+        | fitted (boolean): indicates whether the model has been fitted or not.
+        | fixed_effects (array): the estimated fixed effects of the regression model.
+        | fe_std (array): the estimated standard errors of the fixed effects.
         | random_effects (array): the estimated area level random effects
-        | re_std (number): the estimated standard error of the random effects. 
-        | error_std (number): the estimated standard error of the unit level residuals. 
-        | convergence (dict): a dictionnary holding the convergence status and the number of 
-        |   iterations from the model fitting algorithm. 
+        | re_std (number): the estimated standard error of the random effects.
+        | error_std (number): the estimated standard error of the unit level residuals.
+        | convergence (dict): a dictionnary holding the convergence status and the number of
+        |   iterations from the model fitting algorithm.
         | goodness (dict): a dictionary holding the log-likelihood, AIC, and BIC.
-        | gamma (dict): ratio of the between-area variability (re_std**2) to the total 
-        |   variability (re_std**2 + error_std**2 / a_factor). 
+        | gamma (dict): ratio of the between-area variability (re_std**2) to the total
+        |   variability (re_std**2 + error_std**2 / a_factor).
 
     Prediction related attributes
-        | areap (array): the list of areas for the prediction. 
-        | number_reps (int): number of replicates for the bootstrap MSE estimation. 
-        | area_est (array): area level EBLUP estimates. 
-        | area_mse (array): area level taylor estimation of the MSE. 
+        | areap (array): the list of areas for the prediction.
+        | number_reps (int): number of replicates for the bootstrap MSE estimation.
+        | area_est (array): area level EBLUP estimates.
+        | area_mse (array): area level taylor estimation of the MSE.
 
     Main methods
         | fit(): fits the linear mixed model to estimate the model parameters using REMl or ML
-        |   methods. 
-        | predict(): predicts the area level indicator estimates which includes both the point 
-        |   estimates and the taylor MSE estimate. 
+        |   methods.
+        | predict(): predicts the area level indicator estimates which includes both the point
+        |   estimates and the taylor MSE estimate.
     """
 
     def __init__(
         self,
         method: str = "MOM",
         boxcox: Optional[float] = None,
-        constant: Number = 0,
+        constant: Optional[Number] = None,
         indicator: Optional[Any] = None,
     ):
 
@@ -119,36 +115,36 @@ class EllUnitModel:
         if self.method not in ("REML", "ML", "MOM"):
             raise AssertionError("Value provided for method is not valid!")
         self.indicator = indicator
-        self.boxcox = {"lambda": boxcox, "constant": constant}
+        self.boxcox: dict[str, Optional[Number]] = {"lambda": boxcox, "constant": constant}
 
         # Sample data
-        self.scales: np.ndarray = np.array([])
-        self.afactors: Dict[Any, float] = {}
-        self.ys: np.ndarray = np.array([])
-        self.Xs: np.ndarray = np.array([])
-        self.areas: np.ndarray = np.array([])
-        self.areas_list: np.ndarray = np.array([])
-        self.samp_size: Dict[Any, int] = {}
-        self.ys_mean: np.ndarray = np.array([])
-        self.Xs_mean: np.ndarray = np.array([])
+        self.scales: np.ndarray
+        self.afactors: DictStrNum
+        self.ys: np.ndarray
+        self.Xs: np.ndarray
+        self.areas: np.ndarray
+        self.areas_list: np.ndarray
+        self.samp_size: DictStrNum
+        self.ys_mean: np.ndarray
+        self.Xs_mean: np.ndarray
 
         # Fitted data
         self.fitted: bool = False
-        self.fixed_effects: np.ndarray = np.array([])
-        self.fe_std: np.ndarray = np.array([])
-        self.random_effects: np.ndarray = np.array([])
-        self.re_std: float = 0
-        self.error_std: float = 0
-        self.convergence: Dict[str, Union[float, int, bool]] = {}
-        self.goodness: Dict[str, float] = {}  # loglikehood, deviance, AIC, BIC
-        self.gamma: Dict[Any, float] = {}
+        self.fixed_effects: np.ndarray
+        self.fe_std: np.ndarray
+        self.random_effects: np.ndarray
+        self.re_std: float
+        self.error_std: float
+        self.convergence: dict[str, Union[float, int, bool]] = {}
+        self.goodness: dict[str, Number] = {}  # loglikehood, deviance, AIC, BIC
+        self.gamma: DictStrNum
 
         # Predict(ion/ed) data
-        self.areap: np.ndarray = np.array([])
-        self.Xp_mean: np.ndarray = np.array([])
-        self.number_reps: int = 0
-        self.area_est: Dict[Any, float] = {}
-        self.area_mse: Dict[Any, float] = {}
+        self.areap: np.ndarray
+        self.Xp_mean: np.ndarray
+        self.number_reps: int
+        self.area_est: DictStrNum
+        self.area_mse: DictStrNum
 
     def fit(
         self,
@@ -163,18 +159,18 @@ class EllUnitModel:
     ) -> None:
         """Fits the linear mixed models to estimate the model parameters that is the fixed
         effects, the random effects standard error and the unit level residuals' standard error.
-        In addition, the method provides statistics related to the model fitting e.g. convergence 
+        In addition, the method provides statistics related to the model fitting e.g. convergence
         status, log-likelihood, AIC, BIC, and more.
 
         Args:
-            ys (Array):  an array of the output sample observations. 
+            ys (Array):  an array of the output sample observations.
             Xs (Array): an multi-dimensional array of the sample auxiliary information.
-            areas (Array): provides the area of the sampled observations. 
-            samp_weight (Optional[Array], optional): An array of the sample weights. 
+            areas (Array): provides the area of the sampled observations.
+            samp_weight (Optional[Array], optional): An array of the sample weights.
                 Defaults to None.
-            scales (Union[Array, Number], optional): the scale factor for the unit level errors. 
+            scales (Union[Array, Number], optional): the scale factor for the unit level errors.
                 If a single number of provided, the same number will be applied to all observations. Defaults to 1.
-            intercept (bool, optional): An boolean to indicate whether an intercept need to be 
+            intercept (bool, optional): An boolean to indicate whether an intercept need to be
                 added to Xs. Defaults to True.
         """
 
@@ -187,16 +183,19 @@ class EllUnitModel:
                 Xs = np.insert(Xs.reshape(n, 1), 0, 1, axis=1)
             else:
                 Xs = np.insert(Xs, 0, 1, axis=1)
-        if samp_weight is not None and isinstance(samp_weight, pd.DataFrame):
+        if samp_weight is not None:
             samp_weight = formats.numpy_array(samp_weight)
+
         if isinstance(scales, (float, int)):
-            scales = np.ones(ys.shape[0]) * scales
+            scales = np.asarray(np.ones(ys.shape[0]) * scales)
         else:
             scales = formats.numpy_array(scales)
 
         if self.method in ("REML", "ML"):
             eb_ul = EbUnitModel(
-                method=self.method, boxcox=self.boxcox["lambda"], constant=self.boxcox["constant"],
+                method=self.method,
+                boxcox=self.boxcox["lambda"],
+                constant=self.boxcox["constant"],
             )
             eb_ul.fit(ys, Xs, areas, samp_weight, scales, False, tol=tol, maxiter=maxiter)
             self.scales = eb_ul.scales
@@ -222,7 +221,7 @@ class EllUnitModel:
                 ys, llambda=self.boxcox["lambda"], constant=self.boxcox["constant"], inverse=False
             )
             ols_fit = sm.OLS(ys_transformed, Xs).fit()
-            re_ols = basic_functions.averageby(areas, ols_fit.resid)
+            # re_ols = basic_functions.averageby(areas, ols_fit.resid)
             self.fixed_effects = ols_fit.params
             self.scales = scales
             self.ys = ys
@@ -248,16 +247,16 @@ class EllUnitModel:
         max_array_length: int,
         show_progress: bool,
         **kwargs: Any,
-    ) -> Tuple[np.ndarray, np.ndarray]:
+    ) -> tuple[np.ndarray, np.ndarray]:
 
         areas = np.unique(area)
         nb_areas = len(areas)
         if show_progress:
-            k = 0
             bar_length = min(50, nb_areas)
             steps = np.linspace(1, nb_areas - 1, bar_length).astype(int)
             print(f"Generating the {number_samples} replicates samples")
 
+        k = 0
         eta = np.zeros((number_samples, nb_areas)) * np.nan
         for i, d in enumerate(areas):
             aread = area == d
@@ -289,7 +288,10 @@ class EllUnitModel:
                     )
 
             z_d = basic_functions.transform(
-                y_d, llambda=self.boxcox["lambda"], constant=self.boxcox["constant"], inverse=True,
+                y_d,
+                llambda=self.boxcox["lambda"],
+                constant=self.boxcox["constant"],
+                inverse=True,
             )
             eta[:, i] = np.apply_along_axis(indicator, axis=1, arr=z_d, **kwargs)
 
@@ -299,29 +301,30 @@ class EllUnitModel:
         ell_estimate = np.mean(eta, axis=0)
         ell_mse = np.mean(np.power(eta - ell_estimate[None, :], 2), axis=0)
 
-        return ell_estimate, ell_mse
+        return np.asarray(ell_estimate), np.asarray(ell_mse)
 
     def _predict_indicator_nonparametric(
         self,
         number_samples: int,
-        indicator: Callable[..., np.ndarray],
+        indicator: Callable[..., Array],
         mu: np.ndarray,
         area: np.ndarray,
         total_residuals: np.ndarray,
         max_array_length: int,
         show_progress: bool,
         **kwargs: Any,
-    ) -> Tuple[np.ndarray, np.ndarray]:
+    ) -> tuple[np.ndarray, np.ndarray]:
 
         areas = np.unique(area)
         nb_areas = len(areas)
         if show_progress:
-            k = 0
             bar_length = min(50, nb_areas)
             steps = np.linspace(1, nb_areas - 1, bar_length).astype(int)
             print(f"Generating the {number_samples} replicates samples")
 
+        k = 0
         area_effects = basic_functions.averageby(self.areas, total_residuals)
+        unit_errors = None
         for i, d in enumerate(self.areas_list):
             total_residuals_d = total_residuals[self.areas == d]
             if i == 0:
@@ -338,6 +341,7 @@ class EllUnitModel:
             number_cycles = int(number_samples // cycle_size)
             last_cycle_size = number_samples % cycle_size
 
+            y_d = None
             for j in range(number_cycles + 1):
                 if j == number_cycles:
                     cycle_size = last_cycle_size
@@ -364,7 +368,7 @@ class EllUnitModel:
         ell_estimate = np.mean(eta, axis=0)
         ell_mse = np.mean(np.power(eta - ell_estimate[None, :], 2), axis=0)
 
-        return ell_estimate, ell_mse
+        return np.asarray(ell_estimate), np.asarray(ell_mse)
 
     def predict(
         self,
@@ -378,23 +382,23 @@ class EllUnitModel:
         show_progress: bool = True,
         **kwargs: Any,
     ) -> None:
-        """Predicts the area level indicator and its the MSE estimates. 
+        """Predicts the area level indicator and its the MSE estimates.
 
         Args:
             X (Array): an multi-dimensional array of the auxiliary variables for the population.
             area (Array): provides the area of the population units.
             indicator (Callable[..., Array]): a user defined function which computes the area level
-                indicators. The function should take y (output variable) as the first parameters, 
-                additional parameters can be used. Use ***kwargs* to transfer the additional 
+                indicators. The function should take y (output variable) as the first parameters,
+                additional parameters can be used. Use ***kwargs* to transfer the additional
                 parameters.
             number_samples (int): [description]
             scale (Array, optional): [description]. Defaults to 1.
-            intercept (bool, optional): An boolean to indicate whether an intercept need to be 
+            intercept (bool, optional): An boolean to indicate whether an intercept need to be
                 added to X. Defaults to True.
-            max_array_length (int, optional): controls the number of replicates to generate at 
-                the same time. This parameter helps with performance. The number can be reduce or 
+            max_array_length (int, optional): controls the number of replicates to generate at
+                the same time. This parameter helps with performance. The number can be reduce or
                 increase based on the user's computer RAM capacity. Defaults to int(100e6).
-            show_progress (bool, optional): shows a bar progress of the MC replicates 
+            show_progress (bool, optional): shows a bar progress of the MC replicates
                 calculations. Defaults to True.
 
         Raises:
@@ -406,6 +410,7 @@ class EllUnitModel:
                 "The model must be fitted first with .fit() before running the prediction."
             )
 
+        X = formats.numpy_array(X)
         self.number_samples = int(number_samples)
         if isinstance(scale, (float, int)):
             scale = np.ones(X.shape[0]) * scale
